#include "stdafx.h"
#include "MyTypes.h"
#include "bitmap.h"
#include "mem.h"
#include "error.h"

void BITMAP_Init(TBitmap *p_pBitmap, u16 p_width, u16 p_height, u8 p_format)
{
	p_pBitmap->width = p_width;
	p_pBitmap->height = p_height;
	p_pBitmap->format = p_format;
	p_pBitmap->pMem = NULL;
	p_pBitmap->pPalette = NULL;
}

u32 BITMAP_Alloc(TBitmap *p_pBitmap, u16 p_width, u16 p_height, u8 p_format)
{
	u32	bpp;

	BITMAP_Init(p_pBitmap, p_width, p_height, p_format);

	switch(p_format)
	{
		case BITMAP_256:
		case BITMAP_GREY:
			bpp = 1;
			break;

		case BITMAP_RGB:
		case BITMAP_BGR:
			bpp = 3;
			break;
	}

	if ((p_pBitmap->pMem = MEM_Alloc(p_width * p_height * bpp)) == NULL)
	{
		return (FALSE);
	}

	if (p_format == BITMAP_256)
	{
		if ((p_pBitmap->pPalette = MEM_Alloc(256 * 3)) == NULL)
		{
			MEM_Free(&p_pBitmap->pMem);
			return (FALSE);
		}
	}
	else
	{
		p_pBitmap->pPalette = NULL;
	}

	return (TRUE);
}

void BITMAP_Free(TBitmap *p_pBitmap)
{
	MEM_Free(&p_pBitmap->pMem);
	MEM_Free(&p_pBitmap->pPalette);
}

void BITMAP_Blt(TBitmap *p_pSrcbitmap, TBitmap *p_pDstbitmap, s32 posx, s32 posy)
{
	u32		i;
	s32		x, y, srcx, srcy, dstx, dsty, width, height, a, b;
	u8		*pSrcptr;
	u8		*pDstptr;

	srcx = 0;
	srcy = 0;
	dstx = posx;
	dsty = posy;
	width = p_pSrcbitmap->width;
	height = p_pSrcbitmap->height;

	/// clip
	if (dstx + width <= 0 || dsty+height <= 0 || dstx >= p_pDstbitmap->width || dsty >= p_pDstbitmap->height)
	{
		return;
	}

	if (dstx < 0)
	{
		width += dstx;
		srcx = -dstx;
		dstx = 0;
	}

	if (dsty < 0)
	{
		height += dsty;
		srcy = -dsty;
		dsty = 0;
	}

	if (dstx + width > p_pDstbitmap->width)
	{
		width = p_pDstbitmap->width - dstx;
	}

	if (dsty + height > p_pDstbitmap->height)
	{
		height = p_pDstbitmap->height - dsty;
	}

	/// draw
	if (p_pDstbitmap->format == BITMAP_256 || p_pDstbitmap->format == BITMAP_GREY)
	{
		Error("BITMAP_Blt: format not supported\n");
		return;
	}

	if (p_pSrcbitmap->format == BITMAP_256 || p_pSrcbitmap->format == BITMAP_GREY)
	{
		/// 8 -> 24
		u8	color[3];

		a = 1;
		b = 0;
		if (p_pDstbitmap->format == BITMAP_BGR)
		{
			a = -1;
			b = 2;
		}

		for (y=0; y<height; y++)
		{
			pSrcptr = BITMAP_ADDR8(p_pSrcbitmap, srcx, srcy + y);
			pDstptr = BITMAP_ADDR24(p_pDstbitmap, dstx, dsty + y);

			for (x=0; x<width; x++)
			{
				if (p_pSrcbitmap->format == BITMAP_256)
				{
					color[0] = p_pSrcbitmap->pPalette[*pSrcptr * 3];
					color[1] = p_pSrcbitmap->pPalette[*pSrcptr * 3 + 1];
					color[2] = p_pSrcbitmap->pPalette[*pSrcptr * 3 + 2];
				}
				else
				{
					color[0] = color[1] = color[2] = *pSrcptr;
				}
				pSrcptr++;

				for (i=0; i<3; i++)
				{
					pDstptr[a*i+b] = color[i];
				}
				pDstptr += 3;
			}
		}
	}
	else
	{
		/// 24 -> 24
		a = 1;
		b = 0;
		if (p_pDstbitmap->format != p_pSrcbitmap->format)
		{
			a = -1;
			b = 2;
		}

		for (y=0; y<height; y++)
		{
			pSrcptr = BITMAP_ADDR24(p_pSrcbitmap, srcx, srcy + y);
			pDstptr = BITMAP_ADDR24(p_pDstbitmap, dstx, dsty + y);

			for (x=0; x<width; x++)
			{
				for (i=0; i<3; i++)
				{
					pDstptr[a*i+b] = pSrcptr[i];
				}
				pSrcptr += 3;
				pDstptr += 3;
			}
		}
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
///
/// lit un point
///
/////////////////////////////////////////////////////////////////////////////////////////////////////
COLORREF BITMAP_GetPix(TBitmap *p_bm, s32 p_x, s32 p_y)
{
	if (p_x >= 0 && p_x < p_bm->width &&
		p_y >= 0 && p_y < p_bm->height)
	{
		u8	*pPtr;

		switch (p_bm->format)
		{
			case BITMAP_BGR:
				pPtr = BITMAP_ADDR24(p_bm, p_x, p_y);
				return ((pPtr[0] << 16) + (pPtr[1] << 8) + pPtr[2]);

			case BITMAP_RGB:
				pPtr = BITMAP_ADDR24(p_bm, p_x, p_y);
				return ((pPtr[2] << 16) + (pPtr[1] << 8) + pPtr[0]);

			case BITMAP_GREY:
				pPtr = BITMAP_ADDR8(p_bm, p_x, p_y);
				return ((pPtr[0] << 16) + (pPtr[0] << 8) + pPtr[0]);

			default:
				return (0);
		}
	}
	else
	{
		return (0);
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
///
/// crit un point
///
/////////////////////////////////////////////////////////////////////////////////////////////////////
void BITMAP_SetPix(TBitmap *p_bm, s32 p_x, s32 p_y, COLORREF p_coul)
{
	if (p_x >= 0 && p_x < p_bm->width &&
		p_y >= 0 && p_y < p_bm->height)
	{
		u8	*pPtr;

		switch (p_bm->format)
		{
			case BITMAP_BGR:
				pPtr = BITMAP_ADDR24(p_bm, p_x, p_y);
				*pPtr++ = (u8)(p_coul >> 16);
				*pPtr++ = (u8)(p_coul >> 8);
				*pPtr   = (u8)p_coul;
				break;

			case BITMAP_RGB:
				pPtr = BITMAP_ADDR24(p_bm, p_x, p_y);
				*pPtr++ = (u8)p_coul;
				*pPtr++ = (u8)(p_coul >> 8);
				*pPtr   = (u8)(p_coul >> 16);
				break;

			case BITMAP_GREY:
				pPtr = BITMAP_ADDR8(p_bm, p_x, p_y);
				*pPtr   = (u8)p_coul;
				break;
		}
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
///
/// dessine un rectangle vide
///
/////////////////////////////////////////////////////////////////////////////////////////////////////
void BITMAP_Rect(TBitmap *p_bm, u32 p_x, u32 p_y, u32 p_x2, u32 p_y2, COLORREF p_coul)
{
	u32	i;

	for (i=p_x; i<=p_x2; i++)
	{
		BITMAP_SetPix(p_bm, i, p_y, p_coul);
		BITMAP_SetPix(p_bm, i, p_y2, p_coul);
	}

	for (i=p_y; i<=p_y2; i++)
	{
		BITMAP_SetPix(p_bm, p_x, i, p_coul);
		BITMAP_SetPix(p_bm, p_x2, i, p_coul);
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
///
/// dessine un rectangle plein
///
/////////////////////////////////////////////////////////////////////////////////////////////////////
void BITMAP_FillRect(TBitmap *p_bm, u32 p_x, u32 p_y, u32 p_x2, u32 p_y2, COLORREF p_coul)
{
	u32	x, y;

	for (y=p_y; y<=p_y2; y++)
	{
		for (x=p_x; x<=p_x2; x++)
		{
			BITMAP_SetPix(p_bm, x, y, p_coul);
		}
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
///
/// efface l'cran
///
/////////////////////////////////////////////////////////////////////////////////////////////////////
void BITMAP_Cls(TBitmap *p_bm, COLORREF p_coul)
{
	BITMAP_FillRect(p_bm, 0, 0, p_bm->width, p_bm->height, p_coul);
}

/// todo: faire un blt avec transparence
