////////////////////////////////////////////////////////////////
///
/// TGA load/save routines
///
/////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <string.h>
#include "file.h"
#include "bitmap.h"
#include "error.h"
#include "tga.h"

////////////////////////////////////////////////////
///
/// convertit le header dans le format du fichier
///
////////////////////////////////////////////////////

void TGA_EncodeHeader(TTGA_Header *pHeader, u8 *pHeadBuf)
{
	pHeadBuf[0] = pHeader->alpha;
	pHeadBuf[1] = pHeader->haspalette;
	pHeadBuf[2] = (pHeader->compression << 3) + pHeader->format;
	pHeadBuf[3] = (u8)pHeader->lut;
	pHeadBuf[4] = (u8)(pHeader->lut >> 8);
	pHeadBuf[5] = (u8)pHeader->palsize;
	pHeadBuf[6] = (u8)(pHeader->palsize >> 8);
	pHeadBuf[7] = pHeader->palbits;
	pHeadBuf[8] = (u8)pHeader->unknown;
	pHeadBuf[9] = (u8)(pHeader->unknown >> 8);
	pHeadBuf[10] = (u8)(pHeader->unknown >> 16);
	pHeadBuf[11] = (u8)(pHeader->unknown >> 24);
	pHeadBuf[12] = (u8)pHeader->width;
	pHeadBuf[13] = (u8)(pHeader->width >> 8);
	pHeadBuf[14] = (u8)pHeader->height;
	pHeadBuf[15] = (u8)(pHeader->height >> 8);
	pHeadBuf[16] = pHeader->imgbits;
	pHeadBuf[17] = pHeader->flags;
}

////////////////////////////////////////////////////
///
/// convertit le header dans un format lisible
///
////////////////////////////////////////////////////

void TGA_DecodeHeader(TTGA_Header *pHeader, u8 *pHeadBuf)
{
	pHeader->alpha		= pHeadBuf[0];
	pHeader->format		= pHeadBuf[2] & 3;
	pHeader->compression= pHeadBuf[2] >> 3;
	pHeader->haspalette	= pHeadBuf[1];
	pHeader->lut		= pHeadBuf[3] + (pHeadBuf[4] << 8);
	pHeader->palsize	= pHeadBuf[5] + (pHeadBuf[6] << 8);
	pHeader->palbits	= pHeadBuf[7];
	pHeader->unknown	= pHeadBuf[8] + (pHeadBuf[9] << 8) + (pHeadBuf[10] << 16) + (pHeadBuf[11] << 24);
	pHeader->width		= pHeadBuf[12] + (pHeadBuf[13] << 8);
	pHeader->height		= pHeadBuf[14] + (pHeadBuf[15] << 8);
	pHeader->imgbits	= pHeadBuf[16];
	pHeader->flags		= pHeadBuf[17];
}

////////////////////////////////////////////////////
///
/// sauve un fichier TGA
///
////////////////////////////////////////////////////
u32 TGA_Save(char *p_pFilename, TBitmap *p_pBitmap)
{
	FILE		*pFile;
	u32			i, bpp;
	TTGA_Header	pHeader;
	u8			pHeadBuf[TGA_HEADER_SIZE];

	if (*p_pFilename == 0)
	{
		Error("Error: trying to save image with no name\n");
		return (FALSE);
	}

	if (p_pBitmap->width == 0 || p_pBitmap->height == 0)
	{
		Error("%s: trying to save image with height or width == 0\n", p_pFilename);
		return (FALSE);
	}

	/// cre le header
	switch (p_pBitmap->format)
	{
		case BITMAP_256:
			if (p_pBitmap->pPalette == NULL)
			{
				Error("%s: trying to save image with no palette\n", p_pFilename);
				return (FALSE);
			}
			bpp = 1;
			pHeader.format = TGA_FORMAT_INDEXED;
			pHeader.haspalette = 1;
			pHeader.palsize = 256;
			pHeader.palbits = 24;
			pHeader.imgbits = 8;
			break;

		case BITMAP_GREY:
			bpp = 1;
			pHeader.format = TGA_FORMAT_GREY;
			pHeader.haspalette = 0;
			pHeader.palsize = 0;
			pHeader.palbits = 0;
			pHeader.imgbits = 8;
			break;

		case BITMAP_RGB:
			Error("%s: TGA save mode unsupported\n", p_pFilename);
			break;

		case BITMAP_BGR:
			bpp = 3;
			pHeader.format = TGA_FORMAT_RGB;
			pHeader.haspalette = 0;
			pHeader.palsize = 0;
			pHeader.palbits = 0;
			pHeader.imgbits = 24;
			break;
	}

	pHeader.alpha = 0;
	pHeader.compression = 0;
	pHeader.lut = 0;
	pHeader.unknown = 0;
	pHeader.width = p_pBitmap->width;
	pHeader.height = p_pBitmap->height;
	pHeader.flags = TGA_FLAG_TOP2BOTTOM;
	TGA_EncodeHeader(&pHeader, pHeadBuf);

	/// ouvre le fichier de sortie
	if ((pFile = FILE_Open(p_pFilename, "wb")) == NULL)
	{
		return (FALSE);
	}

	/// crit le header
	if (FILE_Write(&pFile, pHeadBuf, TGA_HEADER_SIZE, p_pFilename) == FALSE)
	{
		return (FALSE);
	}

	/// crit la palette
	if (p_pBitmap->format == BITMAP_256)
	{
		for (i=0; i<256; i++)
		{
			if (FILE_Write(&pFile, &p_pBitmap->pPalette[i * 3 + 2], 1, p_pFilename) == FALSE)
			{
				return (FALSE);
			}

			if (FILE_Write(&pFile, &p_pBitmap->pPalette[i * 3 + 1], 1, p_pFilename) == FALSE)
			{
				return (FALSE);
			}

			if (FILE_Write(&pFile, &p_pBitmap->pPalette[i * 3], 1, p_pFilename) == FALSE)
			{
				return (FALSE);
			}
		}
	}

	/// crit le fichier de sortie
	if (FILE_Write(&pFile, p_pBitmap->pMem, p_pBitmap->width * p_pBitmap->height * bpp, p_pFilename) == FALSE)
	{
		return (FALSE);
	}

	/// ferme le fichier
	FILE_Close(&pFile);

	return (TRUE);
}


#define TGA_ERROR	BITMAP_Free(p_pBitmap); \
					FILE_Close(&pFile); \
					return (FALSE);

////////////////////////////////////////////////////
///
/// lit un fichier TGA
///
////////////////////////////////////////////////////

u32 TGA_Load(char *p_pFilename, TBitmap *p_pBitmap)
{
	TTGA_Header	pHeader;
	u8			pHeadBuf[TGA_HEADER_SIZE];
	FILE		*pFile;
	u8			bmformat;
	u32			memsize, i, j, k, bpp;

	BITMAP_Init(p_pBitmap, 0, 0, 0);

	if ((pFile = FILE_Open(p_pFilename, "rb")) == NULL)
	{
		TGA_ERROR;
	}

	/// lecture du header
	if (FILE_Read(&pFile, pHeadBuf, TGA_HEADER_SIZE, p_pFilename) == FALSE)
	{
		TGA_ERROR;
	}
	TGA_DecodeHeader(&pHeader, pHeadBuf);

	/// vrification du format
	if (pHeader.alpha != 0)
	{
		Error("%s: alpha not supported\n", p_pFilename);
		TGA_ERROR;
	}

	if (pHeader.lut != 0)
	{
		Error("%s: LookUpTable not supported\n", p_pFilename);
		TGA_ERROR;
	}

	if (pHeader.imgbits != 8 && pHeader.imgbits != 24)
	{
		Error("%s: Only 8 & 24 bits images supported\n", p_pFilename);
		TGA_ERROR;
	}

	switch (pHeader.format)
	{
		case TGA_FORMAT_UNKNOWN:
			Error("%s: Unknown format\n", p_pFilename);
			TGA_ERROR;

		case TGA_FORMAT_INDEXED:
			bmformat = BITMAP_256;

			if (pHeader.haspalette == 0)
			{
				Error("%s: Indexed image with no palette\n", p_pFilename);
				TGA_ERROR;
			}

			if (pHeader.palsize > 256)
			{
				Error("%s: Palette with more than 256 colors\n", p_pFilename);
				TGA_ERROR;
			}

			if (pHeader.palbits != 24)
			{
				Error("%s: Palette not in 24 bits\n", p_pFilename);
				TGA_ERROR;
			}
			break;

		case TGA_FORMAT_RGB:
			bmformat = BITMAP_BGR;

			if (pHeader.haspalette == 1)
			{
				Error("%s: TrueColor image with palette\n", p_pFilename);
				TGA_ERROR;
			}
			break;

		case TGA_FORMAT_GREY:
			bmformat = BITMAP_GREY;

			if (pHeader.haspalette == 1)
			{
				Error("%s: Grey image with palette\n", p_pFilename);
				TGA_ERROR;
			}
			break;
	}

	if ((pHeader.flags & 0xC0) != 0)
	{
		Error("%s: Interlaced images not supported\n", p_pFilename);
		TGA_ERROR;
	}

	if (BITMAP_Alloc(p_pBitmap, pHeader.width, pHeader.height, bmformat) == FALSE)
	{
		TGA_ERROR;
	}

	/// lit la palette pour une image indexe
	if (pHeader.format == TGA_FORMAT_INDEXED)							
	{
		for (i=0; i<pHeader.palsize; i++)
		{
			if (FILE_Read(&pFile, pHeadBuf, 3, p_pFilename) == FALSE)
			{
				Error("%s: could not read palette\n", p_pFilename);
				TGA_ERROR;
			}
			p_pBitmap->pPalette[i*3]		= pHeadBuf[2];
			p_pBitmap->pPalette[i*3 + 1]	= pHeadBuf[1];
			p_pBitmap->pPalette[i*3 + 2]	= pHeadBuf[0];
		}
	}

	/// bytes per pixel
	bpp = (pHeader.format == TGA_FORMAT_RGB ? 3 : 1);

	memsize = pHeader.width * pHeader.height * bpp;

	if (pHeader.compression == 0)
	{
		/// lecture des donnes
		if (FILE_Read(&pFile, p_pBitmap->pMem, memsize, p_pFilename) == FALSE)
		{
			TGA_ERROR;
		}
	}
	else
	{
		/// dcompression RLE
		u8	n;

		j = 0;
		while(j < memsize)
		{
			if (FILE_Read(&pFile, pHeadBuf, 1, p_pFilename) == FALSE)
			{
				TGA_ERROR;
			}
			n = pHeadBuf[0];
			if (n < 0x80)
			{
				if (j + bpp * (n+1) > memsize)
				{
					Error("%s: malformed file\n", p_pFilename);
					TGA_ERROR;
				}

				if (FILE_Read(&pFile, &p_pBitmap->pMem[j], bpp * (n+1), p_pFilename) == FALSE)
				{
					TGA_ERROR;
				}

				j += bpp * (n+1);
			}
			else
			{
				n -= 127;

				if (j + bpp * n > memsize)
				{
					Error("%s: malformed file\n", p_pFilename);
					TGA_ERROR;
				}

				if (FILE_Read(&pFile, pHeadBuf, bpp, p_pFilename) == FALSE)
				{
					TGA_ERROR;
				}

				for (i=0; i<n; i++)
				{
					for (k=0; k<bpp; k++)
					{
						p_pBitmap->pMem[j++] = pHeadBuf[k];
					}
				}
			}
		}
	}

	if ((pHeader.flags & TGA_FLAG_TOP2BOTTOM) == 0)
	{
		u8	t;
		u32	w = pHeader.width * bpp;
		u32	h = pHeader.height;

		for (i=0; i<h/2; i++)
		{
			for (j=0; j<w; j++)
			{
				t = p_pBitmap->pMem[i * w + j];
				p_pBitmap->pMem[i * w + j] = p_pBitmap->pMem[(h - i - 1) * w + j];
				p_pBitmap->pMem[(h - i - 1) * w + j] = t;
			}
		}
	}

	return (TRUE);
}
