#include "stdafx.h"
#include "MyTypes.h"
#include "bitmap.h"
#include "mem.h"
#include "error.h"

void BITMAP_Init(TBitmap *p_pBitmap, u16 p_width, u16 p_height, u8 p_format)
{
	p_pBitmap->width = p_width;
	p_pBitmap->height = p_height;
	p_pBitmap->format = p_format;
	p_pBitmap->pMem = NULL;
	p_pBitmap->pPalette = NULL;
}

u32 BITMAP_Alloc(TBitmap *p_pBitmap, u16 p_width, u16 p_height, u8 p_format)
{
	u32	bpp;

	BITMAP_Init(p_pBitmap, p_width, p_height, p_format);

	switch(p_format)
	{
		case BITMAP_256:
		case BITMAP_GREY:
			bpp = 1;
			break;

		case BITMAP_RGB:
		case BITMAP_BGR:
			bpp = 3;
			break;
	}

	if ((p_pBitmap->pMem = (u8 *)MEM_Alloc(p_width * p_height * bpp)) == NULL)
	{
		return (FALSE);
	}

	if (p_format == BITMAP_256)
	{
		if ((p_pBitmap->pPalette = (u8 *)MEM_Alloc(256 * 3)) == NULL)
		{
			MEM_Free((void **)&p_pBitmap->pMem);
			return (FALSE);
		}
	}
	else
	{
		p_pBitmap->pPalette = NULL;
	}

	return (TRUE);
}

void BITMAP_Free(TBitmap *p_pBitmap)
{
	MEM_Free((void **)&p_pBitmap->pMem);
	MEM_Free((void **)&p_pBitmap->pPalette);
}

void BITMAP_Blt(TBitmap *p_pSrcbitmap, TBitmap *p_pDstbitmap, s32 posx, s32 posy)
{
	u32		i;
	s32		x, y, srcx, srcy, dstx, dsty, width, height, a, b;
	u8		*pSrcptr;
	u8		*pDstptr;

	srcx = 0;
	srcy = 0;
	dstx = posx;
	dsty = posy;
	width = p_pSrcbitmap->width;
	height = p_pSrcbitmap->height;

	/// clip
	if (dstx + width <= 0 || dsty+height <= 0 || dstx >= p_pDstbitmap->width || dsty >= p_pDstbitmap->height)
	{
		return;
	}

	if (dstx < 0)
	{
		width += dstx;
		srcx = -dstx;
		dstx = 0;
	}

	if (dsty < 0)
	{
		height += dsty;
		srcy = -dsty;
		dsty = 0;
	}

	if (dstx + width > p_pDstbitmap->width)
	{
		width = p_pDstbitmap->width - dstx;
	}

	if (dsty + height > p_pDstbitmap->height)
	{
		height = p_pDstbitmap->height - dsty;
	}

	/// draw
	if (p_pDstbitmap->format == BITMAP_256)
	{
		Error("BITMAP_Blt: format not supported\n");
		return;
	}

	if (p_pSrcbitmap->format == BITMAP_256 || p_pSrcbitmap->format == BITMAP_GREY)
	{
		/// 8 -> 24 / grey
		u8	color[3];

		a = 1;
		b = 0;
		if (p_pDstbitmap->format == BITMAP_BGR)
		{
			a = -1;
			b = 2;
		}

		for (y=0; y<height; y++)
		{
			pSrcptr = BITMAP_ADDR8(p_pSrcbitmap, srcx, srcy + y);

			if (p_pDstbitmap->format == BITMAP_GREY)
			{
				pDstptr = BITMAP_ADDR8(p_pDstbitmap, dstx, dsty + y);
			}
			else
			{
				pDstptr = BITMAP_ADDR24(p_pDstbitmap, dstx, dsty + y);
			}

			for (x=0; x<width; x++)
			{
				if (p_pSrcbitmap->format == BITMAP_256)
				{
					color[0] = p_pSrcbitmap->pPalette[*pSrcptr * 3];
					color[1] = p_pSrcbitmap->pPalette[*pSrcptr * 3 + 1];
					color[2] = p_pSrcbitmap->pPalette[*pSrcptr * 3 + 2];
				}
				else
				{
					color[0] = color[1] = color[2] = *pSrcptr;
				}
				pSrcptr++;

				if (p_pDstbitmap->format == BITMAP_GREY)
				{
					*pDstptr++ = (color[0] + color[1] + color[2])/3;
				}
				else
				{
					for (i=0; i<3; i++)
					{
						pDstptr[a*i+b] = color[i];
					}
					pDstptr += 3;
				}
			}
		}
	}
	else if (p_pDstbitmap->format == BITMAP_GREY)
	{
		/// 24 -> grey

		for (y=0; y<height; y++)
		{
			pSrcptr = BITMAP_ADDR24(p_pSrcbitmap, srcx, srcy + y);
			pDstptr = BITMAP_ADDR8(p_pDstbitmap, dstx, dsty + y);

			for (x=0; x<width; x++)
			{
				*pDstptr++ = (pSrcptr[0] + pSrcptr[1] + pSrcptr[2]) / 3;
				pSrcptr += 3;
			}
		}
	}
	else
	{
		/// 24 -> 24
		a = 1;
		b = 0;
		if (p_pDstbitmap->format != p_pSrcbitmap->format)
		{
			a = -1;
			b = 2;
		}

		for (y=0; y<height; y++)
		{
			pSrcptr = BITMAP_ADDR24(p_pSrcbitmap, srcx, srcy + y);
			pDstptr = BITMAP_ADDR24(p_pDstbitmap, dstx, dsty + y);

			for (x=0; x<width; x++)
			{
				for (i=0; i<3; i++)
				{
					pDstptr[a*i+b] = pSrcptr[i];
				}
				pSrcptr += 3;
				pDstptr += 3;
			}
		}
	}
}

u32 BITMAP_Clr(TBitmap *p_pBitmap, u8 p_r, u8 p_g, u8 p_b)
{
	u32	x, y;

	if (p_pBitmap->pMem == NULL)
	{
		Error("BITMAP_Clr: no memory for bitmap\n");
		return(FALSE);
	}

	for (y=0; y<p_pBitmap->height; y++)
	{
		for (x=0; x<p_pBitmap->width; x++)
		{
			switch(p_pBitmap->format)
			{
				case BITMAP_256:
				case BITMAP_GREY:
					p_pBitmap->pMem[y*p_pBitmap->width + x] = p_r;
					break;

				case BITMAP_RGB:
					p_pBitmap->pMem[(y*p_pBitmap->width + x)*3] = p_r;
					p_pBitmap->pMem[(y*p_pBitmap->width + x)*3+1] = p_g;
					p_pBitmap->pMem[(y*p_pBitmap->width + x)*3+1] = p_b;
					break;

				case BITMAP_BGR:
					p_pBitmap->pMem[(y*p_pBitmap->width + x)*3] = p_b;
					p_pBitmap->pMem[(y*p_pBitmap->width + x)*3+1] = p_g;
					p_pBitmap->pMem[(y*p_pBitmap->width + x)*3+1] = p_r;
					break;
			}
		}
	}

	return (TRUE);
}
